==> kuzu-0.0.11.tar.gz: sdist/kuzu-source/src/processor/operator/persistent/copy_node.cpp <==
#include "processor/operator/persistent/copy_node.h"

#include "common/exception/copy.h"
#include "common/exception/message.h"
#include "common/string_format.h"
#include "storage/stats/nodes_store_statistics.h"
#include "storage/store/string_column_chunk.h"

using namespace kuzu::catalog;
using namespace kuzu::common;
using namespace kuzu::storage;

namespace kuzu {
namespace processor {

CopyNodeSharedState::CopyNodeSharedState(uint64_t& numRows, NodeTableSchema* tableSchema,
    NodeTable* table, MemoryManager* memoryManager, bool isCopyRdf)
    : numRows{numRows}, tableSchema{tableSchema}, table{table}, pkColumnID{0}, hasLoggedWAL{false},
      currentNodeGroupIdx{0}, isCopyRdf{isCopyRdf} {
    auto ftTableSchema = std::make_unique<FactorizedTableSchema>();
    ftTableSchema->appendColumn(
        std::make_unique<ColumnSchema>(false /* flat */, 0 /* dataChunkPos */,
            LogicalTypeUtils::getRowLayoutSize(LogicalType{LogicalTypeID::STRING})));
    fTable = std::make_shared<FactorizedTable>(memoryManager, std::move(ftTableSchema));
}

void CopyNodeSharedState::initializePrimaryKey(const std::string& directory) {
    if (tableSchema->getPrimaryKey()->getDataType()->getLogicalTypeID() != LogicalTypeID::SERIAL) {
        pkIndex = std::make_unique<PrimaryKeyIndexBuilder>(
            StorageUtils::getNodeIndexFName(directory, tableSchema->tableID, DBFileType::ORIGINAL),
            *tableSchema->getPrimaryKey()->getDataType());
        // Since hashIndexBuilder doesn't support dynamic rehash, we need to reserve enough number
        // of slots when copying turtle files.
        pkIndex->bulkReserve(isCopyRdf ? numRows * 3 : numRows);
    }
    for (auto& property : tableSchema->properties) {
        if (property->getPropertyID() == tableSchema->getPrimaryKey()->getPropertyID()) {
            break;
        }
        pkColumnID++;
    }
}

void CopyNodeSharedState::logCopyNodeWALRecord(WAL* wal) {
    std::unique_lock xLck{mtx};
    if (!hasLoggedWAL) {
        wal->logCopyNodeRecord(table->getTableID(), table->getDataFH()->getNumPages());
        wal->flushAllPages();
        hasLoggedWAL = true;
    }
}

CopyNode::CopyNode(std::shared_ptr<CopyNodeSharedState> sharedState, CopyNodeInfo copyNodeInfo,
    std::unique_ptr<ResultSetDescriptor> resultSetDescriptor,
    std::unique_ptr<PhysicalOperator> child, uint32_t id, const std::string& paramsString)
    : Sink{std::move(resultSetDescriptor), PhysicalOperatorType::COPY_NODE, std::move(child), id,
          paramsString},
      sharedState{std::move(sharedState)}, copyNodeInfo{std::move(copyNodeInfo)} {}

void CopyNodeSharedState::appendLocalNodeGroup(std::unique_ptr<NodeGroup> localNodeGroup) {
    std::unique_lock xLck{mtx};
    if (!sharedNodeGroup) {
        sharedNodeGroup = std::move(localNodeGroup);
        return;
    }
    auto numNodesAppended =
        sharedNodeGroup->append(localNodeGroup.get(), 0 /* offsetInNodeGroup */);
    if (sharedNodeGroup->isFull()) {
        auto nodeGroupIdx = getNextNodeGroupIdxWithoutLock();
        CopyNode::writeAndResetNodeGroup(
            nodeGroupIdx, pkIndex.get(), pkColumnID, table, sharedNodeGroup.get(), isCopyRdf);
    }
    if (numNodesAppended < localNodeGroup->getNumNodes()) {
        sharedNodeGroup->append(localNodeGroup.get(), numNodesAppended);
    }
}

void CopyNode::initGlobalStateInternal(ExecutionContext* context) {
    if (!isCopyAllowed()) {
        throw CopyException(ExceptionMessage::notAllowCopyOnNonEmptyTableException());
    }
    sharedState->initialize(copyNodeInfo.wal->getDirectory());
}

void CopyNode::executeInternal(ExecutionContext* context) {
    // CopyNode goes through UNDO log, should be logged and flushed to WAL before making changes.
    sharedState->logCopyNodeWALRecord(copyNodeInfo.wal);
    while (children[0]->getNextTuple(context)) {
        auto originalSelVector =
            resultSet->getDataChunk(copyNodeInfo.dataColumnPoses[0].dataChunkPos)->state->selVector;
        if (sharedState->isCopyRdf) {
            for (auto& dataPos : copyNodeInfo.dataColumnPoses) {
                // All tuples in the resultSet are in the same data chunk.
                auto vectorToAppend = resultSet->getValueVector(dataPos).get();
                vectorToAppend->state->selVector = originalSelVector;
                appendUniqueValueToPKIndex(sharedState->pkIndex.get(), vectorToAppend);
                copyToNodeGroup({dataPos});
            }
        } else {
            copyToNodeGroup(copyNodeInfo.dataColumnPoses);
        }
        resultSet->getDataChunk(copyNodeInfo.dataColumnPoses[0].dataChunkPos)->state->selVector =
            std::move(originalSelVector);
    }
    if (localNodeGroup->getNumNodes() > 0) {
        sharedState->appendLocalNodeGroup(std::move(localNodeGroup));
    }
}

void CopyNode::sliceDataChunk(
    const DataChunk& dataChunk, const std::vector<DataPos>& dataColumnPoses, offset_t offset) {
    auto slicedSelVector = std::make_unique<SelectionVector>(DEFAULT_VECTOR_CAPACITY);
    slicedSelVector->resetSelectorToValuePosBufferWithSize(
        dataChunk.state->selVector->selectedSize - offset);
    for (auto i = 0u; i < slicedSelVector->selectedSize; i++) {
        slicedSelVector->selectedPositions[i] =
            dataChunk.state->selVector->selectedPositions[i + offset];
    }
    dataChunk.state->selVector = std::move(slicedSelVector);
}

void CopyNode::writeAndResetNodeGroup(node_group_idx_t nodeGroupIdx,
    PrimaryKeyIndexBuilder* pkIndex, column_id_t pkColumnID, NodeTable* table, NodeGroup* nodeGroup,
    bool isCopyTurtle) {
    nodeGroup->setNodeGroupIdx(nodeGroupIdx);
    auto startOffset = StorageUtils::getStartOffsetOfNodeGroup(nodeGroupIdx);
    if (pkIndex && !isCopyTurtle) {
        populatePKIndex(pkIndex, nodeGroup->getColumnChunk(pkColumnID), startOffset,
            nodeGroup->getNumNodes() /* startPageIdx */);
    }
    table->append(nodeGroup);
    nodeGroup->resetToEmpty();
}

void CopyNode::populatePKIndex(
    PrimaryKeyIndexBuilder* pkIndex, ColumnChunk* chunk, offset_t startOffset, offset_t numNodes) {
    checkNonNullConstraint(chunk->getNullChunk(), numNodes);
    std::string errorPKValueStr;
    pkIndex->lock();
    try {
        switch (chunk->getDataType().getPhysicalType()) {
        case PhysicalTypeID::INT64: {
            auto numAppendedNodes = appendToPKIndex<int64_t>(pkIndex, chunk, startOffset, numNodes);
            if (numAppendedNodes < numNodes) {
                errorPKValueStr =
                    std::to_string(chunk->getValue<int64_t>(startOffset + numAppendedNodes));
            }
        } break;
        case PhysicalTypeID::STRING: {
            auto numAppendedNodes =
                appendToPKIndex<ku_string_t>(pkIndex, chunk, startOffset, numNodes);
            if (numAppendedNodes < numNodes) {
                errorPKValueStr =
                    chunk->getValue<ku_string_t>(startOffset + numAppendedNodes).getAsString();
            }
        } break;
        default: {
            throw CopyException(ExceptionMessage::invalidPKType(
                LogicalTypeUtils::dataTypeToString(chunk->getDataType())));
        }
        }
    } catch (Exception& e) {
        pkIndex->unlock();
        throw;
    }
    pkIndex->unlock();
    if (!errorPKValueStr.empty()) {
        throw CopyException(ExceptionMessage::existedPKException(errorPKValueStr));
    }
}

void CopyNode::checkNonNullConstraint(NullColumnChunk* nullChunk, offset_t numNodes) {
    for (auto posInChunk = 0u; posInChunk < numNodes; posInChunk++) {
        if (nullChunk->isNull(posInChunk)) {
            throw CopyException(ExceptionMessage::nullPKException());
        }
    }
}

void CopyNode::finalize(ExecutionContext* context) {
    uint64_t numNodes = StorageUtils::getStartOffsetOfNodeGroup(sharedState->getCurNodeGroupIdx());
    if (sharedState->sharedNodeGroup) {
        numNodes += sharedState->sharedNodeGroup->getNumNodes();
        auto nodeGroupIdx = sharedState->getNextNodeGroupIdx();
        writeAndResetNodeGroup(nodeGroupIdx, sharedState->pkIndex.get(), sharedState->pkColumnID,
            sharedState->table, sharedState->sharedNodeGroup.get(), sharedState->isCopyRdf);
    }
    if (sharedState->pkIndex) {
        sharedState->pkIndex->flush();
    }
    std::unordered_set<table_id_t> connectedRelTableIDs;
    connectedRelTableIDs.insert(sharedState->tableSchema->getFwdRelTableIDSet().begin(),
        sharedState->tableSchema->getFwdRelTableIDSet().end());
    connectedRelTableIDs.insert(sharedState->tableSchema->getBwdRelTableIDSet().begin(),
        sharedState->tableSchema->getBwdRelTableIDSet().end());
    for (auto relTableID : connectedRelTableIDs) {
        copyNodeInfo.relsStore->getRelTable(relTableID)
            ->batchInitEmptyRelsForNewNodes(relTableID, numNodes);
    }
    sharedState->table->getNodeStatisticsAndDeletedIDs()->setNumTuplesForTable(
        sharedState->table->getTableID(), numNodes);
    auto outputMsg = stringFormat("{} number of tuples has been copied to table: {}.", numNodes,
        sharedState->tableSchema->tableName.c_str());
    FactorizedTableUtils::appendStringToTable(
        sharedState->fTable.get(), outputMsg, context->memoryManager);
}

bool CopyNode::isCopyAllowed() const {
    auto nodesStatistics = copyNodeInfo.table->getNodeStatisticsAndDeletedIDs();
    return nodesStatistics->getNodeStatisticsAndDeletedIDs(copyNodeInfo.table->getTableID())
               ->getNumTuples() == 0;
}

template<>
uint64_t CopyNode::appendToPKIndex<int64_t>(
    PrimaryKeyIndexBuilder* pkIndex, ColumnChunk* chunk, offset_t startOffset, uint64_t numValues) {
    for (auto i = 0u; i < numValues; i++) {
        auto offset = i + startOffset;
        auto value = chunk->getValue<int64_t>(i);
        if (!pkIndex->append(value, offset)) {
            return i;
        }
    }
    return numValues;
}

template<>
uint64_t CopyNode::appendToPKIndex<ku_string_t>(
    PrimaryKeyIndexBuilder* pkIndex, ColumnChunk* chunk, offset_t startOffset, uint64_t numValues) {
    auto stringColumnChunk = (StringColumnChunk*)chunk;
    for (auto i = 0u; i < numValues; i++) {
        auto offset = i + startOffset;
        auto value = stringColumnChunk->getValue<std::string>(i);
        if (!pkIndex->append(value.c_str(), offset)) {
            return i;
        }
    }
    return numValues;
}

void CopyNode::appendUniqueValueToPKIndex(
    storage::PrimaryKeyIndexBuilder* pkIndex, common::ValueVector* vectorToAppend) {
    auto selVector = std::make_unique<common::SelectionVector>(common::DEFAULT_VECTOR_CAPACITY);
    selVector->resetSelectorToValuePosBuffer();
    common::sel_t nextPos = 0;
    common::offset_t result;
    auto offset = StorageUtils::getStartOffsetOfNodeGroup(sharedState->getCurNodeGroupIdx()) +
                  localNodeGroup->getNumNodes();
    for (auto i = 0u; i < vectorToAppend->state->getNumSelectedValues(); i++) {
        auto uriStr = vectorToAppend->getValue<common::ku_string_t>(i).getAsString();
        if (!pkIndex->lookup(uriStr.c_str(), result)) {
            pkIndex->append(uriStr.c_str(), offset++);
            selVector->selectedPositions[nextPos++] = i;
        }
    }
    selVector->selectedSize = nextPos;
    vectorToAppend->state->selVector = std::move(selVector);
}

void CopyNode::copyToNodeGroup(std::vector<DataPos> dataPoses) {
    auto numAppendedTuples = 0ul;
    auto numTuplesToAppend =
        resultSet->getDataChunk(dataPoses[0].dataChunkPos)->state->getNumSelectedValues();
    while (numAppendedTuples < numTuplesToAppend) {
        auto numAppendedTuplesInNodeGroup =
            localNodeGroup->append(resultSet, dataPoses, numTuplesToAppend - numAppendedTuples);
        numAppendedTuples += numAppendedTuplesInNodeGroup;
        if (localNodeGroup->isFull()) {
            node_group_idx_t nodeGroupIdx;
            nodeGroupIdx = sharedState->getNextNodeGroupIdx();
            writeAndResetNodeGroup(nodeGroupIdx, sharedState->pkIndex.get(),
                sharedState->pkColumnID, sharedState->table, localNodeGroup.get(),
                sharedState->isCopyRdf);
        }
        if (numAppendedTuples < numTuplesToAppend) {
            sliceDataChunk(*resultSet->getDataChunk(copyNodeInfo.dataColumnPoses[0].dataChunkPos),
                copyNodeInfo.dataColumnPoses, (offset_t)numAppendedTuplesInNodeGroup);
        }
    }
}

} // namespace processor
} // namespace kuzu

==> kuzu-0.0.12.tar.gz: sdist/kuzu-source/src/processor/operator/persistent/copy_node.cpp <==
#include "processor/operator/persistent/copy_node.h"

#include "common/exception/copy.h"
#include "common/exception/message.h"
#include "common/string_format.h"
#include "storage/stats/nodes_store_statistics.h"
#include "storage/store/string_column_chunk.h"

using namespace kuzu::catalog;
using namespace kuzu::common;
using namespace kuzu::storage;

namespace kuzu {
namespace processor {

void CopyNodeSharedState::init() {
    if (pkType->getLogicalTypeID() != LogicalTypeID::SERIAL) {
        auto indexFName = StorageUtils::getNodeIndexFName(
            wal->getDirectory(), table->getTableID(), DBFileType::ORIGINAL);
        pkIndex = std::make_unique<PrimaryKeyIndexBuilder>(indexFName, *pkType);
        // Since hashIndexBuilder doesn't support dynamic rehash, we need to reserve enough number
        // of slots when copying turtle files.
        pkIndex->bulkReserve(numRows);
    }
    wal->logCopyNodeRecord(table->getTableID(), table->getDataFH()->getNumPages());
    wal->flushAllPages();
}

void CopyNodeSharedState::appendLocalNodeGroup(std::unique_ptr<NodeGroup> localNodeGroup) {
    std::unique_lock xLck{mtx};
    if (!sharedNodeGroup) {
        sharedNodeGroup = std::move(localNodeGroup);
        return;
    }
    auto numNodesAppended =
        sharedNodeGroup->append(localNodeGroup.get(), 0 /* offsetInNodeGroup */);
    if (sharedNodeGroup->isFull()) {
        auto nodeGroupIdx = getNextNodeGroupIdxWithoutLock();
        CopyNode::writeAndResetNodeGroup(
            nodeGroupIdx, pkIndex.get(), pkColumnIdx, table, sharedNodeGroup.get());
    }
    if (numNodesAppended < localNodeGroup->getNumNodes()) {
        sharedNodeGroup->append(localNodeGroup.get(), numNodesAppended);
    }
}

static bool isEmptyTable(NodeTable* nodeTable) {
    auto nodesStatistics = nodeTable->getNodeStatisticsAndDeletedIDs();
    return nodesStatistics->getNodeStatisticsAndDeletedIDs(nodeTable->getTableID())
               ->getNumTuples() == 0;
}

void CopyNode::initGlobalStateInternal(ExecutionContext* /*context*/) {
    if (!isEmptyTable(info->table)) {
        throw CopyException(ExceptionMessage::notAllowCopyOnNonEmptyTableException());
    }
    sharedState->init();
}

void CopyNode::initLocalStateInternal(ResultSet* resultSet, ExecutionContext* /*context*/) {
    for (auto& pos : info->columnPositions) {
        columnVectors.push_back(resultSet->getValueVector(pos).get());
    }
    localNodeGroup = std::make_unique<storage::NodeGroup>(
        sharedState->columnTypes, sharedState->table->compressionEnabled());
    columnState = resultSet->getDataChunk(info->columnPositions[0].dataChunkPos)->state.get();
}

void CopyNode::executeInternal(ExecutionContext* context) {
    while (children[0]->getNextTuple(context)) {
        auto originalSelVector = columnState->selVector;
        copyToNodeGroup();
        columnState->selVector = std::move(originalSelVector);
    }
    if (localNodeGroup->getNumNodes() > 0) {
        sharedState->appendLocalNodeGroup(std::move(localNodeGroup));
    }
}

void CopyNode::writeAndResetNodeGroup(node_group_idx_t nodeGroupIdx,
    PrimaryKeyIndexBuilder* pkIndex, column_id_t pkColumnID, NodeTable* table,
    NodeGroup* nodeGroup) {
    nodeGroup->setNodeGroupIdx(nodeGroupIdx);
    auto startOffset = StorageUtils::getStartOffsetOfNodeGroup(nodeGroupIdx);
    if (pkIndex) {
        populatePKIndex(pkIndex, nodeGroup->getColumnChunk(pkColumnID), startOffset,
            nodeGroup->getNumNodes() /* startPageIdx */);
    }
    table->append(nodeGroup);
    nodeGroup->resetToEmpty();
}

void CopyNode::populatePKIndex(
    PrimaryKeyIndexBuilder* pkIndex, ColumnChunk* chunk, offset_t startOffset, offset_t numNodes) {
    checkNonNullConstraint(chunk->getNullChunk(), numNodes);
    std::string errorPKValueStr;
    pkIndex->lock();
    try {
        switch (chunk->getDataType().getPhysicalType()) {
        case PhysicalTypeID::INT64: {
            auto numAppendedNodes = appendToPKIndex<int64_t>(pkIndex, chunk, startOffset, numNodes);
            if (numAppendedNodes < numNodes) {
                errorPKValueStr =
                    std::to_string(chunk->getValue<int64_t>(startOffset + numAppendedNodes));
            }
        } break;
        case PhysicalTypeID::STRING: {
            auto numAppendedNodes =
                appendToPKIndex<ku_string_t>(pkIndex, chunk, startOffset, numNodes);
            if (numAppendedNodes < numNodes) {
                errorPKValueStr =
                    chunk->getValue<ku_string_t>(startOffset + numAppendedNodes).getAsString();
            }
        } break;
        default: {
            throw CopyException(ExceptionMessage::invalidPKType(
                LogicalTypeUtils::dataTypeToString(chunk->getDataType())));
        }
        }
    } catch (Exception& e) {
        pkIndex->unlock();
        throw;
    }
    pkIndex->unlock();
    if (!errorPKValueStr.empty()) {
        throw CopyException(ExceptionMessage::existedPKException(errorPKValueStr));
    }
}

void CopyNode::checkNonNullConstraint(NullColumnChunk* nullChunk, offset_t numNodes) {
    for (auto posInChunk = 0u; posInChunk < numNodes; posInChunk++) {
        if (nullChunk->isNull(posInChunk)) {
            throw CopyException(ExceptionMessage::nullPKException());
        }
    }
}

void CopyNode::finalize(ExecutionContext* context) {
    uint64_t numNodes = StorageUtils::getStartOffsetOfNodeGroup(sharedState->getCurNodeGroupIdx());
    if (sharedState->sharedNodeGroup) {
        numNodes += sharedState->sharedNodeGroup->getNumNodes();
        auto nodeGroupIdx = sharedState->getNextNodeGroupIdx();
        writeAndResetNodeGroup(nodeGroupIdx, sharedState->pkIndex.get(), sharedState->pkColumnIdx,
            sharedState->table, sharedState->sharedNodeGroup.get());
    }
    if (sharedState->pkIndex) {
        sharedState->pkIndex->flush();
    }
    for (auto& relTable : info->connectedRelTables) {
        relTable->batchInitEmptyRelsForNewNodes(relTable->getRelTableID(), numNodes);
    }
    sharedState->table->getNodeStatisticsAndDeletedIDs()->setNumTuplesForTable(
        sharedState->table->getTableID(), numNodes);
    auto outputMsg = stringFormat(
        "{} number of tuples has been copied to table: {}.", numNodes, info->tableName.c_str());
    FactorizedTableUtils::appendStringToTable(
        sharedState->fTable.get(), outputMsg, context->memoryManager);
}

template<>
uint64_t CopyNode::appendToPKIndex<int64_t>(
    PrimaryKeyIndexBuilder* pkIndex, ColumnChunk* chunk, offset_t startOffset, uint64_t numValues) {
    for (auto i = 0u; i < numValues; i++) {
        auto offset = i + startOffset;
        auto value = chunk->getValue<int64_t>(i);
        if (!pkIndex->append(value, offset)) {
            return i;
        }
    }
    return numValues;
}

template<>
uint64_t CopyNode::appendToPKIndex<ku_string_t>(
    PrimaryKeyIndexBuilder* pkIndex, ColumnChunk* chunk, offset_t startOffset, uint64_t numValues) {
    auto stringColumnChunk = (StringColumnChunk*)chunk;
    for (auto i = 0u; i < numValues; i++) {
        auto offset = i + startOffset;
        auto value = stringColumnChunk->getValue<std::string>(i);
        if (!pkIndex->append(value.c_str(), offset)) {
            return i;
        }
    }
    return numValues;
}

void CopyNode::copyToNodeGroup() {
    auto numAppendedTuples = 0ul;
    auto numTuplesToAppend = columnState->getNumSelectedValues();
    while (numAppendedTuples < numTuplesToAppend) {
        auto numAppendedTuplesInNodeGroup = localNodeGroup->append(
            columnVectors, columnState, numTuplesToAppend - numAppendedTuples);
        numAppendedTuples += numAppendedTuplesInNodeGroup;
        if (localNodeGroup->isFull()) {
            node_group_idx_t nodeGroupIdx;
            nodeGroupIdx = sharedState->getNextNodeGroupIdx();
            writeAndResetNodeGroup(nodeGroupIdx, sharedState->pkIndex.get(),
                sharedState->pkColumnIdx, sharedState->table, localNodeGroup.get());
        }
        if (numAppendedTuples < numTuplesToAppend) {
            columnState->slice((offset_t)numAppendedTuplesInNodeGroup);
        }
    }
}

} // namespace processor
} // namespace kuzu
